# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

NN_NAME_KEY = "nn_name"
NN_NAME_TEXT = "NN model name"

NN_VERSION_KEY = "nn_version"
NN_VERSION_TEXT = "NN model version"
NN_VERSION_DEFAULT = "default"

NN_INVOKER_KEY = "nn_invoker"
NN_INVOKER_TEXT = "NN invoker"

NN_EXECUTOR_KEY = "nn_executor"
NN_EXECUTOR_TEXT = "NN executor"

NN_DEVICE_KEY = "nn_device"

NN_OPENAI_MODEL_NAME_KEY = NN_NAME_KEY
NN_OPENAI_MODEL_NAME_TEXT = "openai model name"

NN_OPENAI_API_KEY_KEY = "openai_api_key"
NN_OPENAI_API_KEY_TEXT = "openai api key"

NN_OPENAI_API_BASE_KEY = "openai_api_base"
NN_OPENAI_API_BASE_TEXT = "openai api base"

NN_OPENAI_MAX_TOKENS_KEY = "openai_max_tokens"
NN_OPENAI_MAX_TOKENS_TEXT = "openai max tokens"

NN_OPENAI_ORGANIZATION_KEY = "openai_organization"
NN_OPENAI_ORGANIZATION_TEXT = "openai organization"

NN_OPENAI_GPT4_PREFIX = "gpt-4"
NN_OPENAI_GPT35_PREFIX = "gpt-3.5"
NN_OPENAI_EMBEDDING_PREFIX = "text-embedding"

NN_LOCAL_HF_MODEL_CONFIG_FILE = "config.json"
NN_LOCAL_SENTENCE_TRANSFORMERS_CONFIG_FILE = "config_sentence_transformers.json"

NN_HUB_INFER_URL_KEY = "hub_infer_url"
NN_HUB_INFER_URL_TEXT = "hub inference url"
